/***********************************************************************
 * File: numeral.c
 * 	(Extracted from
 *	   "$Id: numeral.c,v 1.9 2006/07/24 15:27:13 maoyam Exp maoyam $")
 *
 * Abstracts:
 *  This C source implements `numeral()' API.
 *
 * Copyright (C) 2005-2006 Maoyam Tokyo, Japan (mailto:maoyam@mail.goo.ne.jp)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 **********************************************************************/

#include "numeral.h"

/* try to put the name of triad digits */
static int try_to_put_numeral(char *naming
				, char *delimit
				, size_t *sz
				, char **dst
				)
{
	size_t	len;

	if (strlen(delimit) + strlen(naming) > *sz - 1) {
		/* buffer overflow */
		return	NUMERAL_BUFFER_OVERFLOW;
	}
	snprintf(*dst, *sz, "%s%s", delimit, naming);
	*sz	-= (len = strlen(*dst));
	*dst	+= len;
	strcpy(delimit, " ");

	return	NUMERAL_NORMAL_END;
}

/* translate numerals (from zero to 999) */
static int triad(char *src
	, size_t digit
	, char *dst
	, size_t sz
	, int *n
	, char *delimit
	)
{
	static char *hundred_namings[] = {
			  "zero", "one", "two", "three", "four"
			, "five", "six", "seven", "eight", "nine"
			};
	static char *ten_namings[] = {
			  "", "", "twenty", "thirty", "forty", "fifty"
			, "sixty", "seventy", "eighty", "ninety"
			};
	static char *teen_namings[] = {
			  "ten", "eleven", "twelve", "thirteen"
			, "fourteen", "fifteen", "sixteen", "seventeen"
			, "eighteen", "nineteen"
			};
	size_t	i, len, t;
	char	*naming;
	int	result;

	/* get the number */
	for (*n = 0, i = 0; i < digit; ++i) {
		if (!isdigit(src[i])) {
			/* illegal arguments */
			return	NUMERAL_ILLEGAL_ARGS;
		}
		*n	= 10 * *n + src[i] - '0';
	}

	/* test: the number is zero? */
	if (*n == 0) {
		/* test: the first digits describes "zero"? */
		if (*delimit == '\0') {
			/* try to put the "zero" to `dst' string */
			naming	= hundred_namings[0];
			if (strlen(naming) > sz - 1) {
				/* buffer overflow */
				return	NUMERAL_BUFFER_OVERFLOW;
			}
			snprintf(dst, sz, "%s", naming);
		}
		return	NUMERAL_NORMAL_END; /* normal end (zero) */
	}

	/* test: the number is greater than 99? */
	if (*n / 100 > 0) {
		/* try to put "hundred(s)" */
		t	= *n / 100;
		naming	= (t == 1)? "hundred": "hundreds";
		if (strlen(delimit) + strlen(hundred_namings[t])
					+ strlen(naming) + 1 > sz - 1) {
			/* buffer overflow */
			return	NUMERAL_BUFFER_OVERFLOW;
		}
		snprintf(dst, sz, "%s%s %s"
			,	delimit, hundred_namings[t], naming
			);
		sz	-= (len = strlen(dst));
		dst	+= len;
		strcpy(delimit, " ");
	}

	/* test: the number is between 20 and 99? */
	if (*n % 100 > 19) {
		t	= (*n % 100) / 10;
		if ((result = try_to_put_numeral(ten_namings[t]
					, delimit
					, &sz
					, &dst
					)) != NUMERAL_NORMAL_END) {
			return	result; /* buffer over flow */
		}

		/* test: the rest of number exist? */
		if (*n % 10 > 0) {
			/* the number is between 1 and 9 */
			t	= *n % 10;
			if ((result = try_to_put_numeral(
					  hundred_namings[t]
					, delimit
					, &sz
					, &dst
					)) != NUMERAL_NORMAL_END) {
				return	result; /* buffer over flow */
			}
		}
	}
	else if (*n % 100 > 9) {
		/* the number is between 10 and 19 */
		t	= *n % 10;
		if ((result = try_to_put_numeral(teen_namings[t]
					, delimit
					, &sz
					, &dst
					)) != NUMERAL_NORMAL_END) {
			return	result; /* buffer over flow */
		}
	}
	else if (*n % 10 > 0) {
		/* the number is between 1 and 9 */
		t	= *n % 10;
		if ((result = try_to_put_numeral(hundred_namings[t]
					, delimit
					, &sz
					, &dst
					)) != NUMERAL_NORMAL_END) {
			return	result; /* buffer over flow */
		}
	}

	return	NUMERAL_NORMAL_END;
}


/* translate a numeral string to the English or American numeral
 * representation
 */
struct numeral_naming_t {
		size_t	len;	/* as the exponent */
		char	*names;	/* names of the digits */
};

static int translate(struct numeral_naming_t *defs
			, size_t defs_n
			, char **src
			, size_t *srclen
			, char **dst
			, size_t *sz
			, char *delimit
			)
{
	size_t	i, dstlen, src_len;
	int	n, result;
	char	*naming;

	/* try to all namings those are defined in `defs[]'... */
	for (i = 0; i < defs_n; ++i) {
		/* initialize src_sz and n */
		n	= 0;

		/* test: the length of src is greater than this name? */
		if (*srclen > defs[i].len) {
			/* get triad digits length */
			src_len = *srclen - defs[i].len;

			/* translate the three digits */
			result = triad(*src
					, src_len
					, *dst
					, *sz
					, &n
					, delimit
					);
			
			if (result != NUMERAL_NORMAL_END) {
				return	result; /* Failed */
			}

			/* test: the three digits is greater than zero?
			 */
			if (n > 0) {
				/* try to put the name */
				*sz	-= (dstlen = strlen(*dst));
				*dst	+= dstlen;
				naming	= defs[i].names;
				if ((strlen(delimit) + strlen(naming)
					+ ((n > 1)? 1: 0))
							> *sz - 1) {
					/* Failed: buffer overflow */
					return	NUMERAL_BUFFER_OVERFLOW;
				}
				if (defs[i].len > 0) {
					snprintf(*dst, *sz, "%s%s%s"
						, delimit, naming
						, ((n > 1)? "s": "")
						);
					*sz -= (dstlen = strlen(*dst));
					*dst += dstlen;
				}
			}

			/* increment the src pointer */
			*src	+= (*srclen - defs[i].len);
			*srclen	= defs[i].len;
		}
	}

	return	NUMERAL_NORMAL_END;
}

/*
 * This API translates a numeral string to English or American notation.
 *
 * arguments:
 *	src is pointing to the numeral string.
 *	dst is pointing to output string area.
 *	sz indicates the size of `dst' pointing to area.
 *	style indicates the numeral representation style for English
 *	or American.
 *
 * returns:
 *	NUMERAL_NORMAL_END (= 0)	: normal end
 *	NUMERAL_ILLEGAL_ARGS (= 1)	: illegal arguments
 *	NUMERAL_BUFFER_OVERFLOW (= 2)	: buffer overflow
 */
int numeral(char *src
	, char *dst
	, size_t sz
	, NUMERAL_REPRE_STYLE_t style
	)
{
	static struct numeral_naming_t
		  american_defs[] = {
			  {  99, "duotrigintillion"		}
			, {  96, "untrigintillion"		}
			, {  93, "trigintillion"		}
			, {  90, "novemvigintillion"		}
			, {  87, "octovigintillion"		}
			, {  84, "septenvigintillion"		}
			, {  81, "sexvigintillion"		}
			, {  78, "quinvigintillion"		}
			, {  75, "quattuorvigintillion"		}
			, {  72, "trevigintillion"		}
			, {  69, "duovigintillion"		}
			, {  66, "unvigintillion"		}
			, {  63, "vigintillion"			}
			, {  60, "novemdecillion"		}
			, {  57, "octodecillion"		}
			, {  54, "septendecillion"		}
			, {  51, "sexdecillion"			}
			, {  48, "quindecillion"		}
			, {  45, "quattuordecillion"		}
			, {  42, "tredecillion"			}
			, {  39, "duodecillion"			}
			, {  36, "undecillion"			}
			, {  33, "decillion"			}
			, {  30, "nonillion"			}
			, {  27, "octillion"			}
			, {  24, "septillion"			}
			, {  21, "sextillion"			}
			, {  18, "quintillion"			}
			, {  15, "quadrillion"			}
			, {  12, "trillion"			}
			, {   9,  "billion"			}
			, {   6,  "million"			}
			, {   3,  "thousand"			}
			, {   0,  ""				}
			}
		, european_defs[] = {
			  { 195, "duotrigintilliard"		}
			, { 192, "duotrigintillion"		}
			, { 189, "untrigintilliard"		}
			, { 186, "untrigintillion"		}
			, { 183, "trigintilliard"		}
			, { 180, "trigintillion"		}
			, { 177, "novemvigintilliard"		}
			, { 174, "novemvigintillion"		}
			, { 171, "octovigintilliard"		}
			, { 168, "octovigintillion"		}
			, { 165, "septenvigintilliard"		}
			, { 162, "septenvigintillion"		}
			, { 159, "sexvigintilliard"		}
			, { 156, "sexvigintillion"		}
			, { 153, "quinvigintilliard"		}
			, { 150, "quinvigintillion"		}
			, { 147, "quattuorvigintilliard"	}
			, { 144, "quattuorvigintillion"		}
			, { 141, "trevigintilliard"		}
			, { 138, "trevigintillion"		}
			, { 135, "duovigintilliard"		}
			, { 132, "duovigintillion"		}
			, { 129, "unvigintilliard"		}
			, { 126, "unvigintillion"		}
			, { 123, "vigintilliard"		}
			, { 120, "vigintillion"			}
			, { 117, "novemdecilliard"		}
			, { 114, "novemdecillion"		}
			, { 111, "octodecilliard"		}
			, { 108, "octodecillion"		}
			, { 105, "septendecilliard"		}
			, { 102, "septendecillion"		}
			, {  99, "sexdecilliard"		}
			, {  96, "sexdecillion"			}
			, {  93, "quindecilliard"		}
			, {  90, "quindecillion"		}
			, {  87, "quattuordecilliard"		}
			, {  84, "quattuordecillion"		}
			, {  81, "tredecilliard"		}
			, {  78, "tredecillion"			}
			, {  75, "duodecilliard"		}
			, {  72, "duodecillion"			}
			, {  69, "undecilliard"			}
			, {  66, "undecillion"			}
			, {  63, "decilliard"			}
			, {  60, "decillion"			}
			, {  57, "nonilliard"			}
			, {  54, "nonillion"			}
			, {  51, "octilliard"			}
			, {  48, "octillion"			}
			, {  45, "septilliard"			}
			, {  42, "septillion"			}
			, {  39, "sextilliard"			}
			, {  36, "sextillion"			}
			, {  33, "quintilliard"			}
			, {  30, "quintillion"			}
			, {  27, "quadrilliard"			}
			, {  24, "quadrillion"			}
			, {  21, "trilliard"			}
			, {  18, "trillion"			}
			, {  15,  "billiard"			}
			, {  12,  "billion"			}
			, {   9,  "milliard"			}
			, {   6,  "million"			}
			, {   3,  "thousand"			}
			, {   0,  ""				}
			};
	struct numeral_naming_t *defs;
	size_t	srclen, dstlen, defs_n;
	char	delimit[2], *naming;
	int	result;

	/* test: illegal arguments? */
	if (((src == NULL)
	|| dst == NULL)
	|| sz < 2) {
		/* Failed: illegal arguments */
		return	NUMERAL_ILLEGAL_ARGS;
	}

	/* get length of `src' string */
	srclen = strlen(src);

	/* test: the first column is a sign? */
	if (*src == '+') {
		/* skip the positive sign */
		++src, --srclen;
	}
	else if (*src == '-') {
		/* try to put "minus " */
		naming	= "minus ";
		if (strlen(naming) > sz - 1) {
			/* Failed: buffer overflow */
			return	NUMERAL_BUFFER_OVERFLOW;
		}
		snprintf(dst, sz, "%s",	naming);
		sz -= (dstlen = strlen(dst));
		dst += dstlen;

		/* skip the negative sign */
		++src, --srclen;
	}

	/* place defs according to `style' (default is American) */
	if (style == NUMERAL_STYLE_ENGLISH) {
		defs	= european_defs;
		defs_n	= sizeof(european_defs)
					/ sizeof(european_defs[0]);
	}
	else {
		defs	= american_defs;
		defs_n	= sizeof(american_defs)
					/ sizeof(american_defs[0]);
	}
	style	= NUMERAL_STYLE_AMERICAN;

	/* test: the length of src is illegal? */
	if ((srclen == 0) || srclen > defs->len + 3) {
		/* Failed: illegal arguments */
		return	NUMERAL_ILLEGAL_ARGS;
	}

	/* initialize the delimit */
	*delimit	= '\0';

	/* translate the digits to English or American numeral
	 * representation
	 */
	result	= translate(defs
			, defs_n
			, &src
			, &srclen
			, &dst
			, &sz
			, delimit
			);

	return	NUMERAL_NORMAL_END; /* translated the number */
}
